@@ -10,4 +10,11 @@ def floyd_warshall(adj: npBoolMatrix, infinity: int) -> npUInt64Matrix:
     dist[np.diag_indices_from(dist)] = 0
     for k in range(len(dist)):
         np.minimum(dist, dist[:, k, None] + dist[None, k, :], out=dist)
-    return dist
\ No newline at end of file
+    return dist
+
+
+def transitive_closure(leq: npBoolMatrix):
+    n = len(leq)
+    dist = floyd_warshall(leq, infinity=n)
+    rel: npBoolMatrix = (dist < len(dist))
+    return rel
\ No newline at end of file
